from __future__ import annotations

import ckan.plugins.toolkit as tk
import ckanext.scheming.helpers as sch
from ckan.logic import NotFound


def get_relations_info(pkg_type: str) -> list[tuple[str, str, str]]:
    """Return information about relation (related_entity, related_entity_type, relation_type)
    of specified package type (pkg_type) from schema.

    Returns:
        List of tuples of related entities: entity, entity_type, relation_type.
    """
    schema = sch.scheming_get_schema('dataset', pkg_type)
    if not schema:
        return []
    return [(field['related_entity'], field['related_entity_type'], field['relation_type'])
            for field in schema['dataset_fields'] if 'relationship_related_entity' in field.get('validators', '')]


def get_relation_field(pkg_type: str, object_entity: str, object_entity_type: str, relation_type: str) \
        -> dict[str, str]:
    """Return field dict for specified package type (pkg_type) describes relation with specified entity
    (object_entity, object_entity_type) and type of relation (relation_type)."""
    schema = sch.scheming_get_schema('dataset', pkg_type)
    if not schema:
        return []
    for field in schema['dataset_fields']:
        if (
                field.get('related_entity') == object_entity and
                field.get('related_entity_type') == object_entity_type and
                field.get('relation_type') == relation_type
        ):
            return field
    return {}


def entity_name_by_id(entity_id):
    """
    Returns pkg name by its id
    """

    try:
        pkg = tk.get_action("package_show")({"ignore_auth": True}, {"id": entity_id})
        if pkg:
            return pkg.get("name")
    except NotFound:
        pass

    try:
        org = tk.get_action("organization_show")({"ignore_auth": True}, {"id": entity_id})
        if org:
            return org.get("name")
    except NotFound:
        pass

    try:
        group = tk.get_action("group_show")({"ignore_auth": True}, {"id": entity_id})
        if group:
            return group.get("name")
    except NotFound:
        pass




from __future__ import annotations

import ckan.logic as logic
import ckan.plugins.toolkit as tk
import ckanext.relationship.logic.schema as schema
from ckan.logic import validate
from ckanext.relationship.model.relationship import Relationship
from ckanext.relationship.utils import entity_name_by_id
from ckanext.toolbelt.decorators import Collector
from sqlalchemy import or_

NotFound = logic.NotFound

action, get_actions = Collector("relationship").split()


@action
@validate(schema.relation_create)
def relation_create(context, data_dict) -> list[dict[str, str]]:
    """Create relation with specified type (relation_type) between two entities specified
    by ids (subject_id, object_id). Also create reverse relation."""
    tk.check_access('relationship_relation_create', context, data_dict)

    subject_id = data_dict['subject_id']
    object_id = data_dict['object_id']
    relation_type = data_dict.get('relation_type')

    if Relationship.by_object_id(subject_id, object_id, relation_type):
        return

    relation = Relationship(
        subject_id=subject_id,
        object_id=object_id,
        relation_type=relation_type
    )

    reverse_relation = Relationship(
        subject_id=object_id,
        object_id=subject_id,
        relation_type=Relationship.reverse_relation_type[relation_type]
    )

    context['session'].add(relation)
    context['session'].add(reverse_relation)
    context['session'].commit()

    return [rel.as_dict() for rel in (relation, reverse_relation)]


@action
@validate(schema.relation_delete)
def relation_delete(context, data_dict) -> list[dict[str, str]]:
    """Delete relation with specified type (relation_type) between two entities specified
    by ids (subject_id, object_id). Also delete reverse relation."""
    tk.check_access('relationship_relation_delete', context, data_dict)

    subject_id = data_dict['subject_id']
    subject_name = entity_name_by_id(data_dict['subject_id'])
    object_id = data_dict['object_id']
    object_name = entity_name_by_id(data_dict['object_id'])
    relation_type = data_dict.get('relation_type')

    relation = context['session'].query(Relationship) \
        .filter(or_(Relationship.subject_id == subject_id,
                    Relationship.subject_id == subject_name),
                or_(Relationship.object_id == object_id,
                    Relationship.object_id == object_name)
                )

    if relation_type:
        relation = relation.filter(Relationship.relation_type == relation_type)

    relation = relation.all()

    reverse_relation = context['session'].query(Relationship) \
        .filter(or_(Relationship.subject_id == object_id,
                    Relationship.subject_id == object_name),
                or_(Relationship.object_id == subject_id,
                    Relationship.object_id == subject_name),
                )

    if relation_type:
        reverse_relation = reverse_relation.filter(Relationship.relation_type ==
                                                   Relationship.reverse_relation_type[relation_type])

    reverse_relation = reverse_relation.all()

    [context['session'].delete(rel) for rel in relation]
    [context['session'].delete(rel) for rel in reverse_relation]
    context['session'].commit()
    return [rel[0].as_dict() for rel in (relation, reverse_relation) if len(rel) > 0]


@action
@validate(schema.relations_list)
def relations_list(context, data_dict) -> list[dict[str, str]]:
    """Return dicts list of relation of specified entity (object_entity, object_type) related with specified
    type of relation (relation_type) with entity specified by id (subject_id)."""
    tk.check_access('relationship_relations_list', context, data_dict)

    subject_id = data_dict['subject_id']
    object_entity = data_dict.get('object_entity')
    object_entity = 'group' if object_entity and object_entity == 'organization' else object_entity
    object_type = data_dict.get('object_type')
    relation_type = data_dict.get('relation_type')

    relations = Relationship.by_subject_id(subject_id,
                                           object_entity,
                                           object_type,
                                           relation_type)
    if not relations:
        return []
    return [rel.as_dict() for rel in relations]


@action
@validate(schema.relations_ids_list)
def relations_ids_list(context, data_dict) -> list[str]:
    """Return ids list of specified entity (object_entity, object_type) related with specified
    type of relation (relation_type) with entity specified by id (subject_id)."""
    tk.check_access('relationship_relations_ids_list', context, data_dict)

    rel_list = relations_list(context, data_dict)

    return list(set([rel["object_id"] for rel in rel_list]))


@action
@validate(schema.get_entity_list)
def get_entity_list(context, data_dict) -> list[str]:
    """Return ids list of specified entity (entity, entity_type)"""
    tk.check_access('relationship_get_entity_list', context, data_dict)

    model = context['model']

    entity = data_dict['entity']
    entity = entity if entity != 'organization' else 'group'

    entity_type = data_dict['entity_type']

    entity_class = logic.model_name_to_class(model, entity)

    entity_list = (context['session'].query(entity_class.id, entity_class.name, entity_class.title)
                   .filter(entity_class.state != 'deleted')
                   .filter(entity_class.type == entity_type).all())

    return entity_list



import ckan.logic as logic
import ckan.model as model
from ckan.model.types import make_uuid
from sqlalchemy import Column, Text, or_

from .base import Base


class Relationship(Base):
    __tablename__ = 'relationship_relationship'
    id: str = Column(Text, primary_key=True, default=make_uuid)
    subject_id: str = Column(Text, nullable=False)
    object_id: str = Column(Text, nullable=False)
    relation_type: str = Column(Text, nullable=False)

    reverse_relation_type = {
        'related_to': 'related_to',
        'child_of': 'parent_of',
        'parent_of': 'child_of'
    }

    def __repr__(self):
        return (
            'Relationship('
            f'id={self.id!r},'
            f'subject_id={self.subject_id!r},'
            f'object_id={self.object_id!r},'
            f'relation_type={self.relation_type!r},'
            ')'
        )

    def as_dict(self):
        id = self.id
        subject_id = self.subject_id
        object_id = self.object_id
        relation_type = self.relation_type
        return ({'id': id,
                 'subject_id': subject_id,
                 'object_id': object_id,
                 'relation_type': relation_type
                 })

    @classmethod
    def by_object_id(cls, subject_id, object_id, relation_type):
        subject_name = _entity_name_by_id(subject_id)
        object_name = _entity_name_by_id(object_id)

        return model.Session.query(cls). \
            filter(or_(cls.subject_id == subject_id,
                       cls.subject_id == subject_name)). \
            filter(or_(cls.object_id == object_id,
                       cls.object_id == object_name)). \
            filter(cls.relation_type == relation_type).one_or_none()

    @classmethod
    def by_subject_id(cls, subject_id, object_entity, object_type=None, relation_type=None):
        subject_name = _entity_name_by_id(subject_id)

        q = model.Session.query(cls).filter(or_(cls.subject_id == subject_id,
                                                cls.subject_id == subject_name))

        if object_entity:
            object_class = logic.model_name_to_class(model, object_entity)
            q = q.filter(or_(object_class.id == cls.object_id,
                             object_class.name == cls.object_id))

            if object_type:
                q = q.filter(object_class.type == object_type)

        if relation_type:
            q = q.filter(cls.relation_type == relation_type)

        return q.distinct().all()


def _entity_name_by_id(entity_id):
    """
    Returns entity (package or organization or group) name by its id
    """

    pkg = model.Session.query(model.Package).filter(model.Package.id == entity_id).one_or_none()
    if pkg:
        return pkg.name
    group = model.Session.query(model.Group).filter(model.Group.id == entity_id).one_or_none()
    if pkg:
        return group.name



import ckan.plugins as plugins
import ckan.plugins.toolkit as tk
import ckanext.relationship.helpers as helpers
import ckanext.relationship.views as views
import ckanext.relationship.logic.action as action
import ckanext.relationship.logic.auth as auth
import ckanext.relationship.logic.validators as validators
import ckanext.relationship.utils as utils
import ckanext.scheming.helpers as sch
from ckan.lib.search import rebuild
from ckan.logic import NotFound


class RelationshipPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IActions)
    plugins.implements(plugins.IAuthFunctions)
    plugins.implements(plugins.IValidators)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IBlueprint)
    plugins.implements(plugins.IPackageController, inherit=True)

    # IConfigurer
    def update_config(self, config_):
        tk.add_template_directory(config_, 'templates')
        tk.add_public_directory(config_, 'public')
        tk.add_resource('assets', 'relationship')

    # IActions
    def get_actions(self):
        return action.get_actions()

    # IAuthFunctions
    def get_auth_functions(self):
        return auth.get_auth_functions()

    # IValidators
    def get_validators(self):
        return validators.get_validators()

    # ITemplateHelpers
    def get_helpers(self):
        return helpers.get_helpers()

    # IBlueprint
    def get_blueprint(self):
        return views.get_blueprints()

    # IPackageController
    def after_create(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)
        return _update_relations(context, pkg_dict)

    def after_update(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)
        return _update_relations(context, pkg_dict)

    def after_delete(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)

        subject_id = pkg_dict["id"]

        relations_ids_list = tk.get_action('relationship_relations_ids_list')(context, {'subject_id': subject_id})

        for object_id in relations_ids_list:
            tk.get_action('relationship_relation_delete')(context, {'subject_id': subject_id, 'object_id': object_id})

            try:
                rebuild(object_id)
            except NotFound:
                pass
        rebuild(subject_id)

    def before_index(self, pkg_dict):
        pkg_id = pkg_dict['id']
        pkg_type = pkg_dict['type']
        schema = sch.scheming_get_schema('dataset', pkg_type)
        if not schema:
            return pkg_dict
        relations_info = utils.get_relations_info(pkg_type)
        for related_entity, related_entity_type, relation_type in relations_info:
            relations_ids = tk.get_action('relationship_relations_ids_list')({}, {'subject_id': pkg_id,
                                                                                  'object_entity': related_entity,
                                                                                  'object_type': related_entity_type,
                                                                                  'relation_type': relation_type})

            if not relations_ids:
                continue
            field = utils.get_relation_field(pkg_type, related_entity, related_entity_type, relation_type)
            pkg_dict[f'vocab_{field["field_name"]}'] = relations_ids

            del pkg_dict[field["field_name"]]

        return pkg_dict

    def after_show(self, context, pkg_dict):
        pkg_id = pkg_dict['id']
        pkg_type = pkg_dict['type']
        relations_info = utils.get_relations_info(pkg_type)
        for related_entity, related_entity_type, relation_type in relations_info:
            field = utils.get_relation_field(pkg_type, related_entity, related_entity_type, relation_type)
            pkg_dict[field['field_name']] = \
                tk.get_action('relationship_relations_ids_list')(context, {'subject_id': pkg_id,
                                                                           'object_entity': related_entity,
                                                                           'object_type': related_entity_type,
                                                                           'relation_type': relation_type})


def _update_relations(context, pkg_dict):
    subject_id = pkg_dict['id']
    add_relations = pkg_dict.get('add_relations', [])
    del_relations = pkg_dict.get('del_relations', [])
    if not add_relations and not del_relations:
        return pkg_dict
    for object_id, relation_type in del_relations + add_relations:
        if (object_id, relation_type) in add_relations:
            tk.get_action('relationship_relation_create')(context, {'subject_id': subject_id,
                                                                    'object_id': object_id,
                                                                    'relation_type': relation_type})
        else:
            tk.get_action('relationship_relation_delete')(context, {'subject_id': subject_id,
                                                                    'object_id': object_id,
                                                                    'relation_type': relation_type})

        try:
            rebuild(object_id)
        except NotFound:
            pass
    rebuild(subject_id)
    return pkg_dict



import ckan.plugins.toolkit as tk
from flask import Blueprint, jsonify


def get_blueprints():
    return [
        relationships,
    ]


relationships = Blueprint("relationships", __name__)


@relationships.route("/api/2/util/relationships/autocomplete")
def relationships_autocomplete():
    incomplete = tk.request.args.get("incomplete", "")
    entity_type = tk.request.args.get("entity_type", "dataset")
    result = []
    if incomplete:
        packages = tk.get_action('package_search')({}, {
            'q': incomplete,
            'fq': f'type:{entity_type}',
            'fl': 'id, title',
            'rows': 100,
            'include_private': True,
            "sort": 'score desc',
        })['results']

        filtered_packages = filter(
            lambda pkg: incomplete.lower() in pkg["title"].lower(), packages
        )
        result = {
            "ResultSet": {
                "Result": [
                    {
                        "name": pkg["id"],
                        "title": pkg["title"],
                    } for pkg in filtered_packages
                ]
            }
        }
    return jsonify(result)



from __future__ import annotations

from ckan.plugins import toolkit as tk
from ckanext.toolbelt.decorators import Collector

auth, get_auth_functions = Collector("relationship").split()


@auth
def relation_create(context, data_dict):
    return {'success': True}


@auth
def relation_delete(context, data_dict):
    return {'success': True}


@auth
@tk.auth_allow_anonymous_access
def relations_list(context, data_dict):
    return {'success': True}


@auth
@tk.auth_allow_anonymous_access
def relations_ids_list(context, data_dict):
    return {'success': True}


@auth
@tk.auth_allow_anonymous_access
def get_entity_list(context, data_dict):
    return {'success': True}
